==> pyrefly-0.22.2.tar.gz: pyrefly-0.22.2/pyrefly/lib/types/keywords.rs <==
/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

use pyrefly_derive::TypeEq;
use pyrefly_derive::Visit;
use pyrefly_derive::VisitMut;
use pyrefly_util::visit::Visit;
use pyrefly_util::visit::VisitMut;
use ruff_python_ast::name::Name;
use starlark_map::ordered_map::OrderedMap;

use crate::types::callable::FuncMetadata;
use crate::types::callable::FunctionKind;
use crate::types::types::Type;

#[derive(Debug, Clone, PartialEq, Eq, TypeEq, PartialOrd, Ord, Hash)]
pub struct TypeMap(pub OrderedMap<Name, Type>);

impl TypeMap {
    pub fn new() -> Self {
        Self(OrderedMap::new())
    }

    fn get_bool(&self, name: &Name, default: bool) -> bool {
        self.0
            .get(name)
            .and_then(|t| t.as_bool())
            .unwrap_or(default)
    }
}

impl Visit<Type> for TypeMap {
    fn recurse<'a>(&'a self, f: &mut dyn FnMut(&'a Type)) {
        for (_, ty) in self.0.iter() {
            ty.visit(f);
        }
    }
}

impl VisitMut<Type> for TypeMap {
    fn recurse_mut(&mut self, f: &mut dyn FnMut(&mut Type)) {
        for (_, ty) in self.0.iter_mut() {
            ty.visit_mut(f);
        }
    }
}

/// Wraps the result of a function call whose keyword arguments have typing effects.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct KwCall {
    /// The metadata of the called function.
    pub func_metadata: FuncMetadata,
    /// The keyword arguments that the function was called with.
    pub keywords: TypeMap,
    /// The return type of the call.
    pub return_ty: Type,
}

impl KwCall {
    pub fn has_function_kind(&self, kind: FunctionKind) -> bool {
        self.func_metadata.kind == kind
    }
}

/// Parameters to `typing.dataclass_transform`.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#dataclass-transform-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassTransformKeywords {
    pub eq_default: bool,
    pub order_default: bool,
    pub kw_only_default: bool,
    pub frozen_default: bool,
    // TODO(rechen): add field_specifiers
}

impl DataclassTransformKeywords {
    const EQ_DEFAULT: Name = Name::new_static("eq_default");
    const ORDER_DEFAULT: Name = Name::new_static("order_default");
    const KW_ONLY_DEFAULT: Name = Name::new_static("kw_only_default");
    const FROZEN_DEFAULT: Name = Name::new_static("frozen_default");

    pub fn from_type_map(map: &TypeMap) -> Self {
        Self {
            eq_default: map.get_bool(&Self::EQ_DEFAULT, true),
            order_default: map.get_bool(&Self::ORDER_DEFAULT, false),
            kw_only_default: map.get_bool(&Self::KW_ONLY_DEFAULT, false),
            frozen_default: map.get_bool(&Self::FROZEN_DEFAULT, false),
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new())
    }
}

/// Parameters to dataclass field specifiers.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#field-specifier-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassFieldKeywords {
    pub init: bool,
    pub default: bool,
    /// None means that kw_only was not explicitly set
    pub kw_only: Option<bool>,
    // TODO(rechen): add factory, alias, and converter
}

impl DataclassFieldKeywords {
    const INIT: Name = Name::new_static("init");
    /// We combine default and default_factory into a single "default" keyword indicating whether
    /// the field has a default. The default value isn't stored.
    const DEFAULT: Name = Name::new_static("default");
    const DEFAULT_FACTORY: Name = Name::new_static("default_factory");
    const KW_ONLY: Name = Name::new_static("kw_only");

    pub fn from_type_map(map: &TypeMap) -> Self {
        let init = map.get_bool(&Self::INIT, true);
        let default =
            map.0.contains_key(&Self::DEFAULT) | map.0.contains_key(&Self::DEFAULT_FACTORY);
        let kw_only = map.0.get(&Self::KW_ONLY).and_then(|t| t.as_bool());
        Self {
            init,
            default,
            kw_only,
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new())
    }

    pub fn is_kw_only(&self) -> bool {
        self.kw_only == Some(true)
    }
}

/// Dataclass parameters.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#decorator-function-and-class-metaclass-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassKeywords {
    pub init: bool,
    pub order: bool,
    pub frozen: bool,
    pub match_args: bool,
    pub kw_only: bool,
    pub eq: bool,
    pub unsafe_hash: bool,
    // TODO(rechen): add `slots`
}

impl DataclassKeywords {
    const INIT: Name = Name::new_static("init");
    const ORDER: Name = Name::new_static("order");
    const FROZEN: Name = Name::new_static("frozen");
    const MATCH_ARGS: Name = Name::new_static("match_args");
    const KW_ONLY: Name = Name::new_static("kw_only");
    const EQ: Name = Name::new_static("eq");
    const UNSAFE_HASH: Name = Name::new_static("unsafe_hash");

    pub fn from_type_map(map: &TypeMap, defaults: &DataclassTransformKeywords) -> Self {
        Self {
            init: map.get_bool(&Self::INIT, true),
            order: map.get_bool(&Self::ORDER, defaults.order_default),
            frozen: map.get_bool(&Self::FROZEN, defaults.frozen_default),
            match_args: map.get_bool(&Self::MATCH_ARGS, true),
            kw_only: map.get_bool(&Self::KW_ONLY, defaults.kw_only_default),
            eq: map.get_bool(&Self::EQ, defaults.eq_default),
            unsafe_hash: map.get_bool(&Self::UNSAFE_HASH, false),
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new(), &DataclassTransformKeywords::new())
    }
}

==> pyrefly-0.23.1.tar.gz: pyrefly-0.23.1/pyrefly/lib/types/keywords.rs <==
/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

use pyrefly_derive::TypeEq;
use pyrefly_derive::Visit;
use pyrefly_derive::VisitMut;
use pyrefly_util::visit::Visit;
use pyrefly_util::visit::VisitMut;
use ruff_python_ast::name::Name;
use starlark_map::ordered_map::OrderedMap;

use crate::types::callable::FuncMetadata;
use crate::types::callable::FunctionKind;
use crate::types::literal::Lit;
use crate::types::tuple::Tuple;
use crate::types::types::CalleeKind;
use crate::types::types::Type;

#[derive(Debug, Clone, PartialEq, Eq, TypeEq, PartialOrd, Ord, Hash)]
pub struct TypeMap(pub OrderedMap<Name, Type>);

impl TypeMap {
    pub fn new() -> Self {
        Self(OrderedMap::new())
    }

    fn get_bool(&self, name: &Name, default: bool) -> bool {
        self.0
            .get(name)
            .and_then(|t| t.as_bool())
            .unwrap_or(default)
    }

    fn get_string(&self, name: &Name) -> Option<&str> {
        self.0.get(name).and_then(|t| match t {
            Type::Literal(Lit::Str(s)) => Some(&**s),
            _ => None,
        })
    }
}

impl Visit<Type> for TypeMap {
    fn recurse<'a>(&'a self, f: &mut dyn FnMut(&'a Type)) {
        for (_, ty) in self.0.iter() {
            ty.visit(f);
        }
    }
}

impl VisitMut<Type> for TypeMap {
    fn recurse_mut(&mut self, f: &mut dyn FnMut(&mut Type)) {
        for (_, ty) in self.0.iter_mut() {
            ty.visit_mut(f);
        }
    }
}

/// Wraps the result of a function call whose keyword arguments have typing effects.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct KwCall {
    /// The metadata of the called function.
    pub func_metadata: FuncMetadata,
    /// The keyword arguments that the function was called with.
    pub keywords: TypeMap,
    /// The return type of the call.
    pub return_ty: Type,
}

impl KwCall {
    pub fn has_function_kind(&self, kind: FunctionKind) -> bool {
        self.func_metadata.kind == kind
    }
}

/// Parameters to `typing.dataclass_transform`.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#dataclass-transform-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassTransformKeywords {
    pub eq_default: bool,
    pub order_default: bool,
    pub kw_only_default: bool,
    pub frozen_default: bool,
    pub field_specifiers: Vec<CalleeKind>,
}

impl DataclassTransformKeywords {
    const EQ_DEFAULT: Name = Name::new_static("eq_default");
    const ORDER_DEFAULT: Name = Name::new_static("order_default");
    const KW_ONLY_DEFAULT: Name = Name::new_static("kw_only_default");
    const FROZEN_DEFAULT: Name = Name::new_static("frozen_default");
    const FIELD_SPECIFIERS: Name = Name::new_static("field_specifiers");

    pub fn from_type_map(map: &TypeMap) -> Self {
        Self {
            eq_default: map.get_bool(&Self::EQ_DEFAULT, true),
            order_default: map.get_bool(&Self::ORDER_DEFAULT, false),
            kw_only_default: map.get_bool(&Self::KW_ONLY_DEFAULT, false),
            frozen_default: map.get_bool(&Self::FROZEN_DEFAULT, false),
            field_specifiers: match map.0.get(&Self::FIELD_SPECIFIERS) {
                Some(Type::Tuple(Tuple::Concrete(elts))) => {
                    elts.iter().filter_map(|e| e.callee_kind()).collect()
                }
                _ => Vec::new(),
            },
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new())
    }
}

/// Parameters to dataclass field specifiers.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#field-specifier-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassFieldKeywords {
    pub init: bool,
    pub default: bool,
    /// None means that kw_only was not explicitly set
    pub kw_only: Option<bool>,
    /// Alias that is used in `__init__` rather than the field name. None means no alias
    pub alias: Option<Name>,
    // TODO(rechen): add converter
}

impl DataclassFieldKeywords {
    const INIT: Name = Name::new_static("init");
    /// We combine default, default_factory, and factory into a single "default" keyword indicating
    /// whether the field has a default. The default value isn't stored.
    const DEFAULT: Name = Name::new_static("default");
    const DEFAULT_FACTORY: Name = Name::new_static("default_factory");
    const FACTORY: Name = Name::new_static("factory");
    const KW_ONLY: Name = Name::new_static("kw_only");
    const ALIAS: Name = Name::new_static("alias");

    pub fn from_type_map(map: &TypeMap) -> Self {
        let init = map.get_bool(&Self::INIT, true);
        let default = [&Self::DEFAULT, &Self::DEFAULT_FACTORY, &Self::FACTORY]
            .iter()
            .any(|k| map.0.contains_key(*k));
        let kw_only = map.0.get(&Self::KW_ONLY).and_then(|t| t.as_bool());
        let alias = map.get_string(&Self::ALIAS).map(Name::new);
        Self {
            init,
            default,
            kw_only,
            alias,
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new())
    }

    pub fn is_kw_only(&self) -> bool {
        self.kw_only == Some(true)
    }
}

/// Dataclass parameters.
/// See https://typing.python.org/en/latest/spec/dataclasses.html#decorator-function-and-class-metaclass-parameters.
#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[derive(Visit, VisitMut, TypeEq)]
pub struct DataclassKeywords {
    pub init: bool,
    pub order: bool,
    pub frozen: bool,
    pub match_args: bool,
    pub kw_only: bool,
    pub eq: bool,
    pub unsafe_hash: bool,
    pub slots: bool,
}

impl DataclassKeywords {
    const INIT: Name = Name::new_static("init");
    const ORDER: Name = Name::new_static("order");
    const FROZEN: Name = Name::new_static("frozen");
    const MATCH_ARGS: Name = Name::new_static("match_args");
    const KW_ONLY: Name = Name::new_static("kw_only");
    const EQ: Name = Name::new_static("eq");
    const UNSAFE_HASH: Name = Name::new_static("unsafe_hash");
    const SLOTS: Name = Name::new_static("slots");

    pub fn from_type_map(map: &TypeMap, defaults: &DataclassTransformKeywords) -> Self {
        Self {
            init: map.get_bool(&Self::INIT, true),
            order: map.get_bool(&Self::ORDER, defaults.order_default),
            frozen: map.get_bool(&Self::FROZEN, defaults.frozen_default),
            match_args: map.get_bool(&Self::MATCH_ARGS, true),
            kw_only: map.get_bool(&Self::KW_ONLY, defaults.kw_only_default),
            eq: map.get_bool(&Self::EQ, defaults.eq_default),
            unsafe_hash: map.get_bool(&Self::UNSAFE_HASH, false),
            slots: map.get_bool(&Self::SLOTS, false),
        }
    }

    pub fn new() -> Self {
        Self::from_type_map(&TypeMap::new(), &DataclassTransformKeywords::new())
    }
}
